# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER query normalizer.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

from efilter import engine
from efilter import expression


class Normalizer(engine.VisitorEngine):
    """Optimizes the AST for better performance and simpler structure.

    This class follows the visitor pattern. See documentation on VisitorEngine.

    The returned query will be logically equivalent to what was provided but
    transformations will be made to flatten and optimize the structure. This
    engine works by recognizing certain patterns and replacing them with nicer
    ones, eliminating pointless expressions, and so on.

    Examples:
        # Logical expressions are made variadic:
        Intersection("foo", Intersection("bar", "baz")) # becomes:
        Intersection("foo", "bar", "baz")

        # Let-forms are rotated so that the LHS is a Binding when possible:
        Let(
            Let(
                Binding("Process"),
                Binding("parent")),
            Equivalence(
                Binding("name"),
                Literal("init")))
        # Becomes:
        Let(
            Binding("Process"),
            Let(
                Binding("parent"),
                Equivalence(
                    Binding("name"),
                    Literal("init"))))
    """

    def run(self, *args, **kwargs):
        expr = super(Normalizer, self).run(*args, **kwargs)
        return self.query.subquery(expr)

    def visit_Let(self, expr, **kwargs):
        """Rotate repeated let-forms so they cascade on the RHS.

        Basic let-forms should be rotated as follows:
        (let (let x y) (...)) => (let x (let y) (...))

        These are functionally equivalent, but the latter is easier to follow.

        Returns rotated Let instance.
        """
        lhs = self.visit(expr.lhs, **kwargs)
        rhs = self.visit(expr.rhs)

        if (isinstance(lhs, expression.Let)
                and isinstance(lhs.lhs, expression.Binding)):
            lhs_ = lhs.lhs
            rhs = type(expr)(lhs.rhs, rhs)
            lhs = lhs_

        return type(expr)(lhs, rhs)

    def visit_LetAny(self, expr, **kwargs):
        """let-any|let-each forms are not cascaded.

        This is basically a pass-through function.
        """
        lhs = self.visit(expr.lhs, **kwargs)
        rhs = self.visit(expr.rhs, **kwargs)
        return type(expr)(lhs, rhs)

    def visit_LetEach(self, expr, **kwargs):
        return self.visit_LetAny(expr, **kwargs)

    def visit_Expression(self, expr, **_):
        return expr

    def visit_BinaryExpression(self, expr, **kwargs):
        return self.visit_VariadicExpression(expr, **kwargs)

    def visit_VariadicExpression(self, expr, **kwargs):
        """Pass through n-ary expressions, and eliminate empty branches.

        Variadic and binary expressions recursively visit all their children.

        If all children are eliminated then the parent expression is also
        eliminated:

        (& [removed] [removed]) => [removed]

        If only one child is left, it is promoted to replace the parent node:

        (& True) => True
        """
        children = []
        for child in expr.children:
            branch = self.visit(child, **kwargs)
            if branch is None:
                continue

            if type(branch) is type(expr):
                children.extend(branch.children)
            else:
                children.append(branch)

        if len(children) == 0:
            return None

        if len(children) == 1:
            return children[0]

        return type(expr)(*children)


engine.Engine.register_engine(Normalizer, "normalizer")



#!/usr/bin/env python

from efilter import version
from setuptools import find_packages, setup

setup(name="efilter",
      version=version.get_version(),
      description="EFILTER query language",
      license="Apache 2.0",
      author="Adam Sindelar",
      author_email="adam.sindelar@gmail.com",
      url="https://github.com/google/dotty/",
      packages=find_packages(exclude=["tests*"]),
      package_dir={"efilter": "efilter"})



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system."""

import datetime
import numbers

from efilter import dispatch
from efilter import protocol

# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def hashed(x):
    raise NotImplementedError()


class IHashable(protocol.Protocol):
    _protocol_functions = (hashed,)


# Default implementations:

IHashable.implement(for_types=(basestring, numbers.Number, type(None),
                               tuple, frozenset, datetime.datetime),
                    implementations={hashed: hash})



# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
EFILTER frontend base class.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"


import abc


class Frontend(object):
    """Base class representing parsers or generators of the EFILTER AST."""
    __metaclass__ = abc.ABCMeta

    FRONTENDS = {}

    def __init__(self, original, params=None):
        """Create a frontend parser for this dialect.

        Arguments:
            original: The source code of this query. Most often this is a
                string type, but there are exceptions (e.g. lisp)
            params: Some dialects support parametric queries (for safety) -
                if used, pass them as params. This should be a dict for
                keywords or a tuple for positional.
        """
        super(Frontend, self).__init__()
        self.params = params
        self.original = original

    @abc.abstractproperty
    def root(self):
        """The root of the resultant AST.

        Subclasses MUST implement parsing here.
        """

    @classmethod
    def register_frontend(cls, subcls, shorthand=None):
        cls.register(subcls)

        if shorthand is None:
            shorthand = repr(subcls)

        cls.FRONTENDS[shorthand] = subcls

    @classmethod
    def get_frontend(cls, shorthand):
        return cls.FRONTENDS.get(shorthand)



# -*- coding: utf-8 -*-

# EFILTER Forensic Query Language
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""EFILTER abstract type system."""

from efilter import dispatch
from efilter import protocol
from efilter.protocols import hashable

# Declarations:
# pylint: disable=unused-argument


@dispatch.polymorphic
def indices(x):
    """Return a list of keys to represent 'self' in maps.

    It is sometimes desirable to store objects in sets or dicts, which are
    not hashable. There may be two reasons for this:
    1) Mutable objects may not be hashable for obvious reasons.
    2) Hashable objects must satisfy:
       x == y -> hash(x) == hash(y).
       This requirement may be impossible to meet for certain types that
       can compare as multiple types, for example emulated C enumerations,
       which compare as either integers or strings, but can only be hashed
       as one or the other.

    The former is easily addressed by not using mutable types for non-ephemeral
    state. It's the 21st century, people.

    The Indexable interface addresses the latter, by letting the object return
    a collection of keys, here called 'indices', at which it desires to appear
    in any associative collection. The indices must themselves be hashable.

    See ext.indexset for more details about how indices may be used.
    """
    raise NotImplementedError()


class IIndexable(protocol.Protocol):
    _protocol_functions = (indices,)


# Default implementations:

IIndexable.implement(for_type=hashable.IHashable,
                     implementations={indices: lambda x: (x,)})



import unittest

from efilter import expression
from efilter.frontends import lisp


class ParserTest(unittest.TestCase):
    def assertQueryMatches(self, query, expected):
        parser = lisp.Parser(query)
        actual = parser.root
        self.assertEqual(expected, actual)

    def testBasic(self):
        self.assertQueryMatches(
            ("==", ("var", "foo"), "bar"),
            expression.Equivalence(
                expression.Binding("foo"),
                expression.Literal("bar")))
